(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uunix.ml,v 1.24 2014/02/09 17:27:28 fclement Exp $ *)

let u_qual = Upervasives.qualify "Uunix";;

let u_internal_error = Upervasives.internal_error_qual u_qual;;

let spawn_init init_command command bin cout =
  match (!bin, !cout) with
  | (Some ib, Some oc) -> (ib, oc)
  | (Some _, None) | (None, Some _) -> u_internal_error "spawn_init"
  | _ ->
    let (ic, oc) = Unix.open_process command in
    let ib = Scanf.Scanning.from_channel ic in
    bin := Some ib; cout := Some oc;
    init_command (ib, oc);
    (ib, oc)
;;

type command_line = string;;

type process_io_buffers =
  {
   input_buffer : Scanf.Scanning.scanbuf;
   output_buffer : Buffer.t;
   flush_output : unit -> unit;
  }
;;

let spawn spf cmd piob_ref =
  match !piob_ref with
  | Some piob -> piob
  | None ->
      let (ic, oc) = Unix.open_process cmd in
      spf.Udebug.spf_report (Printf.sprintf "command %S spawned" cmd);
      let ib = Scanf.Scanning.from_channel ic in
      let ob = Buffer.create 1024 in
      let flush () = Ubuffer.flush_buffer oc ob in
      let piob =
        { input_buffer = ib; output_buffer = ob; flush_output = flush; } in
      piob_ref := Some piob;
      piob
;;

let std_process_io_buffers =
  let ib = Scanf.Scanning.stdin in
  let ob = Buffer.create 1024 in
  let flush () = Ubuffer.flush_buffer_on_stdout ob in
  { input_buffer = ib; output_buffer = ob; flush_output = flush; }
;;

(* Connected processes stuff. *)

type pid = int;;

let connect_stdio proc (fdin1, fdout1) (fdin2, fdout2) =
  let pid = Unix.fork () in
  match pid with
  | 0 ->
    Unix.close fdout1;
    Unix.close fdin2;
    Unix.dup2 fdin1 Unix.stdin;
    Unix.close fdin1;
    Unix.dup2 fdout2 Unix.stdout;
    Unix.close fdout2;
    proc ()
  | pid -> pid
;;

let connect_bi_directional proc1 proc2 =
  let p1 = Unix.pipe () in
  let p2 = Unix.pipe () in
  let pid1 = connect_stdio proc1 p2 p1 in
  let pid2 = connect_stdio proc2 p1 p2 in
  (pid1, pid2)
;;

let launch prog args () = Unix.execv prog (Array.of_list (prog :: args))
;;

let launch_connected_processes prog1 args1 prog2 args2 =
  connect_bi_directional (launch prog1 args1) (launch prog2 args2)
;;

let connect_stdio_with_err proc (fdin1, fdout1) (fdin2, fdout2) fderr =
  let pid = Unix.fork () in
  match pid with
  | 0 ->
    Unix.close fdout1;
    Unix.close fdin2;
    Unix.dup2 fdin1 Unix.stdin;
    Unix.close fdin1;
    Unix.dup2 fdout2 Unix.stdout;
    Unix.close fdout2;
    Unix.dup2 fderr Unix.stderr;
    Unix.close fderr;
    proc ()
  | pid -> pid
;;

let connect_bi_directional_with_err proc1 fderr1 proc2 fderr2 =
  let p1 = Unix.pipe () in
  let p2 = Unix.pipe () in
  let pid1 = connect_stdio_with_err proc1 p2 p1 fderr1 in
  let pid2 = connect_stdio_with_err proc2 p1 p2 fderr2 in
  (pid1, pid2)
;;

let launch_connected_processes_with_err
    prog1 args1 fderr1 prog2 args2 fderr2 =
  connect_bi_directional_with_err
    (launch prog1 args1) fderr1 (launch prog2 args2) fderr2
;;

let launch_connected_processes_with_err_on_channels
    prog1 args1 oc1 prog2 args2 oc2 =
  let fderr1 = Unix.descr_of_out_channel oc1 in
  let fderr2 = Unix.descr_of_out_channel oc2 in
  launch_connected_processes_with_err
    prog1 args1 fderr1 prog2 args2 fderr2
;;

let wait_for_normal_termination pid =
  let (_pid, status) = Unix.waitpid [] pid in
  match status with
  | Unix.WEXITED 0 -> ()
  | Unix.WEXITED n ->
    prerr_endline
      (Printf.sprintf "Process %d exited with error %d." pid n);
    exit n;
  | Unix.WSIGNALED n ->
    prerr_endline
      (Printf.sprintf "Process %d was killed by signal %d." pid n);
    exit 1;
  | Unix.WSTOPPED n ->
    prerr_endline
      (Printf.sprintf "Process %d was stopped by signal %d." pid n);
    exit 1;
;;
